/*
*******************************************************************************
*
*      COPYRIGHT (C) 2000 BY ERICSSON EUROLAB DEUTSCHLAND GmbH
*      90411 NUERNBERG, GERMANY, Tel Int + 49 911 5217 100
*
*      The program(s) may be used and/or copied only with the
*      written permission from Ericsson or in accordance
*      with the terms and conditions stipulated in the agreement or
*      contract under which the program(s) have been supplied.
*
*******************************************************************************
*
*      File      : ctm_score.c
*      Purpose   : Scoring tool for counting the errors in a tranmsitted
*                  text message. The tool is able to count inserted, 
*                  deleted, and replaced characters as errors. This is 
*                  achieved by determining an appropriate correlation 
*                  between the time scales of the reference text and the
*                  received text, respectively. This correlation is similar
*                  to the "Viterbi Algorithm" or the "Dynamic Programming"
*                  and aims at the search of the optimal path through the
*                  trellis chart.  The reference text (text1) refers to the
*                  horizontal axis of the trellis, while the received text
*                  (text2) refers to the vertical axis. Therefore,  
*                  characters that are deleted in the received text result 
*                  in an horizontal step in the trellis, while insertions 
*                  lead to a vertical transition of the optimal path.
*                                                                   
*                  Example with three errors (one deletion, one insertion, 
*                  and one replacement). The optimal path from the upper 
*                  left corner to the lower right corner has a metric of 3,
*                  which refers to the numer of detected errors:
*
*                      T  h  i  s     i  s     a     t  e  s  t  
*                  --------------------------------------------
*                  T   0  1  2  3  4  5  6  7  8  9 10 11 12 13
*                  i   1  1  1  2  3  4  5  6  7  8  9 10 11 12
*                  s   2  2  2  1  2  3  4  5  6  7  8  9 10 11
*                      3  3  3  2  1  2  3  4  5  6  7  8  9 10
*                  i   4  4  3  3  2  1  2  3  4  5  6  7  8  9
*                  i   5  5  4  4  3  2  2  3  4  5  6  7  8  9
*                  s   6  6  5  4  4  3  2  3  4  5  6  7  7  8
*                      7  7  6  5  4  4  3  2  3  4  5  6  7  8
*                  a   8  8  7  6  5  5  4  3  2  3  4  5  6  7
*                      9  9  8  7  6  6  5  4  3  2  3  4  5  6
*                  x  10 10  9  8  7  7  6  5  4  3  3  4  5  6
*                  e  11 11 10  9  8  8  7  6  5  4  4  3  4  5
*                  s  12 12 11 10  9  9  8  7  6  5  5  4  3  4
*                  t  13 13 12 11 10 10  9  8  7  6  5  5  4  3
*
*******************************************************************************
*
* $Id: $ 
*
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "typedefs.h"

#define CARR_RTN 13 /* ASCII code for Carriage Return */

/***********************************************************************/

void writeHead()
{
  fprintf(stderr, "\n");
  fprintf(stderr, "***************************************************************\n");
  fprintf(stderr, "     Scoring Program for Calculating Character Error Rates\n");
  fprintf(stderr, "***************************************************************\n\n");
}

/***********************************************************************/

void errorMsg(const char* filename)
{
  fprintf(stderr, "use: %s [-v] reference_text received_text score\n\n", filename);
  fprintf(stderr, "Mandatory parameters:\n");
  fprintf(stderr, "  reference_text : filename with reference text\n");
  fprintf(stderr, "  received_text  : filename with received text\n");
  fprintf(stderr, "  score          : name for output text file with score results\n\n");
  fprintf(stderr, "Options:\n");
  fprintf(stderr, "  -v : Indicates verbose output of the complete trellis.\n");
  fprintf(stderr, "       It's not recommended to use this option if your \n");
  fprintf(stderr, "       reference text is longer than 30 characters.\n\n");
  exit(1);
}

/***********************************************************************/

int main(int argc, const char** argv)
{
  FILE  *text1FilePtr  = NULL;
  FILE  *text2FilePtr  = NULL;
  FILE  *outputFilePtr = NULL;
  
  char* text1;         /* String for the refercence text */
  char* text2;         /* String for the received text   */
  const char* text1FileName;
  const char* text2FileName;
  const char* outputFileName;
  char  tmpChar;
  
  Longint lengthText1 = 0;
  Longint lengthText2 = 0;
  
  Longint cnt;
  Longint cntCols;
  Longint cntRows;
  
  Longint  actualMetric;
  Longint* oldMetrics;
  Longint* newMetrics;
  
  Bool verbose = false;
  
  writeHead();
  
  if (argc<4)
    errorMsg(argv[0]);
  
  /* Check, whether option -v has been used */
  if (strcmp(argv[1], "-v")==0)
    {
      if (argc<5)
        errorMsg(argv[0]);
      verbose = true;
      text1FileName  = argv[2];
      text2FileName  = argv[3];
      outputFileName = argv[4];
    }
  else
    {
      verbose = false;
      text1FileName  = argv[1];
      text2FileName  = argv[2];
      outputFileName = argv[3];
    }
  
  /* Determine the length of Text #1 */
  text1FilePtr = fopen(text1FileName, "rb");
  if (text1FilePtr == (FILE*)NULL)
    {
      fprintf(stderr, "Can't open input text file '%s'\n", text1FileName);
      exit(1);
    }
  do
    {
      if (fread(&tmpChar, sizeof(char), 1, text1FilePtr) < 1)
        break;
      /* Don't count CarriageReturn symbols */
      if (tmpChar!=CARR_RTN)
        lengthText1++;
    }
  while (1);
  
  /* Determine the length of Text #2 */
  text2FilePtr = fopen(text2FileName, "rb");
  if (text2FilePtr == (FILE*)NULL)
    {
      fprintf(stderr, "Can't open input text file '%s'\n", text2FileName);
      exit(1);
    }
  do
    {
      if(fread(&tmpChar, sizeof(char), 1, text2FilePtr) < 1)
        break;
      /* Don't count CarriageReturn symbols */
      if (tmpChar!=CARR_RTN)
        lengthText2++;
    }
  while (1);
  
  /* Close and Re-Open the file with Text #1 */
  fclose(text1FilePtr);
  text1FilePtr = fopen(text1FileName, "rb");
  if (text1FilePtr == (FILE*)NULL)
    {
      fprintf(stderr, "Can't open input text file '%s'\n", text1FileName);
      exit(1);
    }
  
  /* Close and Re-Open the file with Text #2 */
  fclose(text2FilePtr);
  text2FilePtr = fopen(text2FileName, "rb");
  if (text2FilePtr == (FILE*)NULL)
    {
      fprintf(stderr, "Can't open input text file '%s'\n", text2FileName);
      exit(1);
    }
  
  outputFilePtr = fopen(outputFileName, "w");
  if (outputFilePtr == (FILE*)NULL)
    {
      fprintf(stderr, "Can't open output file '%s'\n", outputFileName);
      exit(1);
    }
  
  oldMetrics = (Longint*)calloc(lengthText1, sizeof(Longint));
  newMetrics = (Longint*)calloc(lengthText1, sizeof(Longint));
  text1      = (char*)calloc(lengthText1, sizeof(char));
  text2      = (char*)calloc(lengthText2, sizeof(char));
  
  /* Read in complete Text #1 */
  for (cnt=0; cnt<lengthText1; cnt++)
    {
      /* Read in characters until actual character is unequal CARR_RTN */
      do 
        if(fread(&(text1[cnt]), sizeof(char), 1, text1FilePtr) < 1)
          {
            fprintf(stderr, "\nPremature end of text file '%s'\n", 
                    text1FileName);
            exit(1);
          }
      while (text1[cnt]==CARR_RTN);
    }
  
  /* Read in complete Text #2 */
  for (cnt=0; cnt<lengthText2; cnt++)
    {
      /* Read in characters until actual character is unequal CARR_RTN */
      do 
        if(fread(&(text2[cnt]), sizeof(char), 1, text2FilePtr) < 1)
          {
            fprintf(stderr, "\nPremature end of text file '%s'\n", 
                    text2FileName);
            exit(1);
          }
      while (text2[cnt]==CARR_RTN);
    }
  
  /* Both texts reside in vectors text1 and text2 now... */
  
  if (verbose)
    {
      /* Print out Text #1, Text #2, and headline of the trellis */
      fprintf(stdout, "Reference Text:\n\n");
      for (cnt=0; cnt<lengthText1; cnt++)
        fprintf(stdout, "%c", text1[cnt]);
      
      fprintf(stdout, "\n\nReceived Text:\n\n");
      for (cnt=0; cnt<lengthText2; cnt++)
        fprintf(stdout, "%c", text2[cnt]);
      fprintf(stdout, "\n\nTrellis:\n\n  ");
      for (cntCols=0; cntCols<lengthText1; cntCols++)
        fprintf(stdout, "%3c", text1[cntCols]);
    }
  
  /* Initialize the row vector with the metrics */
  for (cntCols=1; cntCols<lengthText1; cntCols++)
    oldMetrics[cntCols] = maxLongint;
  oldMetrics[0] = 0;
  newMetrics[lengthText1-1] = lengthText1;
    
  /* Process the trellis row-by-row. We only have to store the metric */
  /* values of the actual and the previous row in the trellis.        */
  for (cntRows=0; cntRows<lengthText2; cntRows++)
    {
      for (cntCols=0; cntCols<lengthText1; cntCols++)
        {
          if (text1[cntCols]==text2[cntRows])
            actualMetric=0;
          else
            actualMetric=1;
          if (cntCols>0)
            {
              if ((oldMetrics[cntCols-1] <= oldMetrics[cntCols]) &&
                  (oldMetrics[cntCols-1] <= newMetrics[cntCols-1]))
                /* diagonal step in the trellis: increase of the metric     */
                /* depends on actual metric (might be a replaced character) */
                newMetrics[cntCols] = oldMetrics[cntCols-1]+actualMetric;
              else if (oldMetrics[cntCols] <= newMetrics[cntCols-1])
                /* horizontal step in the trellis: deleted character */
                newMetrics[cntCols] = oldMetrics[cntCols]+1;
              else
                /* vertical step in the trellis: inserted character  */
                newMetrics[cntCols] = newMetrics[cntCols-1]+1;
            }
          else
            /* All cells of the first column in the trellis can have */
            /* only one preceding cell (there are no alternatives).  */
            newMetrics[cntCols] = oldMetrics[cntCols]+actualMetric;
        }
      /* Update the vector with the old metrics */
      for (cntCols=0; cntCols<lengthText1; cntCols++)
        oldMetrics[cntCols] = newMetrics[cntCols];
      
      if (verbose)
        {
          fprintf(stdout, "\n%1c ", text2[cntRows]);
          for (cntCols=0; cntCols<lengthText1; cntCols++)
            fprintf(stdout, "%3.1d", newMetrics[cntCols]);
        }
      else
        fprintf(stderr, "\rScoring in progress... %d%% completed.", 
                (100*(cntRows+1))/lengthText2);
    }
  
  fprintf(stdout, "\n");
  fprintf(stdout, "\nNumber of character errors :  %d", 
          newMetrics[lengthText1-1]);
  fprintf(stdout, "\nLength of reference text   :  %d", lengthText1);
  fprintf(stdout, "\nCharacter error rate (in %%): %5.2f\n", 
          100.0*(float)(newMetrics[lengthText1-1])/(float)(lengthText1));
  
  fprintf(outputFilePtr, "Received text filename     :  %s", text2FileName);
  fprintf(outputFilePtr, "\nNumber of character errors :  %d", 
          newMetrics[lengthText1-1]);
  fprintf(outputFilePtr, "\nLength of reference text   :  %d", lengthText1);
  fprintf(outputFilePtr, "\nCharacter error rate (in %%): %5.2f\n", 
          100.0*(float)(newMetrics[lengthText1-1])/(float)(lengthText1));
  
  return 0;
}
